@@ -25,6 +25,36 @@ use wasm_timer::Delay;
 /// once the transaction has enough `confirmations`. The default number of confirmations
 /// is 1, but may be adjusted with the `confirmations` method. If the transaction does not
 /// have enough confirmations or is not mined, the future will stay in the pending state.
+///
+/// # Example
+///
+///```
+/// # use ethers_providers::{Provider, Http};
+/// # use ethers_core::utils::Ganache;
+/// # use std::convert::TryFrom;
+/// use ethers_providers::Middleware;
+/// use ethers_core::types::TransactionRequest;
+///
+/// # #[tokio::main(flavor = "current_thread")]
+/// # async fn main() -> Result<(), Box<dyn std::error::Error>> {
+/// # let ganache = Ganache::new().spawn();
+/// # let client = Provider::<Http>::try_from(ganache.endpoint()).unwrap();
+/// # let accounts = client.get_accounts().await?;
+/// # let from = accounts[0];
+/// # let to = accounts[1];
+/// # let balance_before = client.get_balance(to, None).await?;
+/// let tx = TransactionRequest::new().to(to).value(1000).from(from);
+/// let receipt = client
+///     .send_transaction(tx, None)
+///     .await?                           // PendingTransaction<_>
+///     .log_msg("Pending transfer hash") // print pending tx hash with message
+///     .await?;                          // Result<Option<TransactionReceipt>, _>
+/// # let _ = receipt;
+/// # let balance_after = client.get_balance(to, None).await?;
+/// # assert_eq!(balance_after, balance_before + 1000);
+/// # Ok(())
+/// # }
+/// ```
 #[pin_project]
 pub struct PendingTransaction<'a, P> {
     tx_hash: TxHash,
@@ -32,8 +62,11 @@ pub struct PendingTransaction<'a, P> {
     provider: &'a Provider<P>,
     state: PendingTxState<'a>,
     interval: Box<dyn Stream<Item = ()> + Send + Unpin>,
+    retries_remaining: usize,
 }
 
+const DEFAULT_RETRIES: usize = 3;
+
 impl<'a, P: JsonRpcClient> PendingTransaction<'a, P> {
     /// Creates a new pending transaction poller from a hash and a provider
     pub fn new(tx_hash: TxHash, provider: &'a Provider<P>) -> Self {
@@ -44,6 +77,7 @@ impl<'a, P: JsonRpcClient> PendingTransaction<'a, P> {
             provider,
             state: PendingTxState::InitialDelay(delay),
             interval: Box::new(interval(DEFAULT_POLL_INTERVAL)),
+            retries_remaining: DEFAULT_RETRIES,
         }
     }
 
@@ -55,14 +89,21 @@ impl<'a, P: JsonRpcClient> PendingTransaction<'a, P> {
         self.provider.clone()
     }
 
+    /// Returns the transaction hash of the pending transaction
+    pub fn tx_hash(&self) -> TxHash {
+        self.tx_hash
+    }
+
     /// Sets the number of confirmations for the pending transaction to resolve
     /// to a receipt
+    #[must_use]
     pub fn confirmations(mut self, confs: usize) -> Self {
         self.confirmations = confs;
         self
     }
 
     /// Sets the polling interval
+    #[must_use]
     pub fn interval<T: Into<Duration>>(mut self, duration: T) -> Self {
         let duration = duration.into();
 
@@ -74,6 +115,37 @@ impl<'a, P: JsonRpcClient> PendingTransaction<'a, P> {
 
         self
     }
+
+    /// Set retries
+    #[must_use]
+    pub fn retries(mut self, retries: usize) -> Self {
+        self.retries_remaining = retries;
+        self
+    }
+}
+
+impl<'a, P> PendingTransaction<'a, P> {
+    /// Allows inspecting the content of a pending transaction in a builder-like way to avoid
+    /// more verbose calls, e.g.:
+    /// `let mined = token.transfer(recipient, amt).send().await?.inspect(|tx| println!(".{}",
+    /// *tx)).await?;`
+    pub fn inspect<F>(self, mut f: F) -> Self
+    where
+        F: FnMut(&Self),
+    {
+        f(&self);
+        self
+    }
+
+    /// Logs the pending transaction hash along with a custom message before it.
+    pub fn log_msg<S: std::fmt::Display>(self, msg: S) -> Self {
+        self.inspect(|s| println!("{}: {:?}", msg, **s))
+    }
+
+    /// Logs the pending transaction's hash
+    pub fn log(self) -> Self {
+        self.inspect(|s| println!("Pending hash: {:?}", **s))
+    }
 }
 
 macro_rules! rewake_with_new_state {
@@ -100,7 +172,7 @@ impl<'a, P: JsonRpcClient> Future for PendingTransaction<'a, P> {
 
         match this.state {
             PendingTxState::InitialDelay(fut) => {
-                let _ready = futures_util::ready!(fut.as_mut().poll(ctx));
+                futures_util::ready!(fut.as_mut().poll(ctx));
                 tracing::debug!("Starting to poll pending tx {:?}", *this.tx_hash);
                 let fut = Box::pin(this.provider.get_transaction(*this.tx_hash));
                 rewake_with_new_state!(ctx, this, PendingTxState::GettingTx(fut));
@@ -127,9 +199,14 @@ impl<'a, P: JsonRpcClient> Future for PendingTransaction<'a, P> {
                 let tx_opt = tx_res.unwrap();
                 // If the tx is no longer in the mempool, return Ok(None)
                 if tx_opt.is_none() {
-                    tracing::debug!("Dropped from mempool, pending tx {:?}", *this.tx_hash);
-                    *this.state = PendingTxState::Completed;
-                    return Poll::Ready(Ok(None))
+                    if *this.retries_remaining == 0 {
+                        tracing::debug!("Dropped from mempool, pending tx {:?}", *this.tx_hash);
+                        *this.state = PendingTxState::Completed;
+                        return Poll::Ready(Ok(None))
+                    }
+
+                    *this.retries_remaining -= 1;
+                    rewake_with_new_state!(ctx, this, PendingTxState::PausedGettingTx);
                 }
 
                 // If it hasn't confirmed yet, poll again later
