@@ -1,4 +1,5 @@
 mod common;
+pub use common::Authorization;
 
 // only used with WS
 #[cfg(feature = "ws")]
@@ -9,6 +10,8 @@ macro_rules! if_wasm {
     )*}
 }
 
+// only used with WS
+#[cfg(feature = "ws")]
 macro_rules! if_not_wasm {
     ($($item:item)*) => {$(
         #[cfg(not(target_arch = "wasm32"))]
@@ -16,15 +19,13 @@ macro_rules! if_not_wasm {
     )*}
 }
 
-if_not_wasm! {
-    #[cfg(feature = "ipc")]
-    mod ipc;
-    #[cfg(feature = "ipc")]
-    pub use ipc::Ipc;
-}
+#[cfg(all(target_family = "unix", feature = "ipc"))]
+mod ipc;
+#[cfg(all(target_family = "unix", feature = "ipc"))]
+pub use ipc::{Ipc, IpcError};
 
 mod http;
-pub use http::{ClientError as HttpClientError, Provider as Http};
+pub use self::http::{ClientError as HttpClientError, Provider as Http};
 
 #[cfg(feature = "ws")]
 mod ws;
@@ -33,7 +34,10 @@ pub use ws::{ClientError as WsClientError, Ws};
 
 mod quorum;
 pub(crate) use quorum::JsonRpcClientWrapper;
-pub use quorum::{Quorum, QuorumProvider, WeightedProvider};
+pub use quorum::{Quorum, QuorumError, QuorumProvider, WeightedProvider};
+
+mod rw;
+pub use rw::{RwClient, RwClientError};
 
 mod mock;
 pub use mock::{MockError, MockProvider};
