@@ -1,6 +1,6 @@
 use crate::{
     provider::ProviderError,
-    transports::common::{JsonRpcError, Notification, Request, Response},
+    transports::common::{JsonRpcError, Request},
     JsonRpcClient, PubsubClient,
 };
 use ethers_core::types::U256;
@@ -12,6 +12,7 @@ use futures_util::{
     stream::{Fuse, Stream, StreamExt},
 };
 use serde::{de::DeserializeOwned, Serialize};
+use serde_json::value::RawValue;
 use std::{
     collections::{btree_map::Entry, BTreeMap},
     fmt::{self, Debug},
@@ -22,6 +23,8 @@ use std::{
 };
 use thiserror::Error;
 
+use super::common::{Params, Response};
+
 if_wasm! {
     use wasm_bindgen::prelude::*;
     use wasm_bindgen_futures::spawn_local;
@@ -59,11 +62,14 @@ if_not_wasm! {
     type Message = tungstenite::protocol::Message;
     type WsError = tungstenite::Error;
     type WsStreamItem = Result<Message, WsError>;
+    use super::Authorization;
     use tracing::{debug, error, warn};
+    use http::Request as HttpRequest;
+    use tungstenite::client::IntoClientRequest;
 }
 
-type Pending = oneshot::Sender<Result<serde_json::Value, JsonRpcError>>;
-type Subscription = mpsc::UnboundedSender<serde_json::Value>;
+type Pending = oneshot::Sender<Result<Box<RawValue>, JsonRpcError>>;
+type Subscription = mpsc::UnboundedSender<Box<RawValue>>;
 
 /// Instructions for the `WsServer`.
 enum Instruction {
@@ -75,13 +81,6 @@ enum Instruction {
     Unsubscribe { id: U256 },
 }
 
-#[derive(Debug, serde::Deserialize)]
-#[serde(untagged)]
-enum Incoming {
-    Notification(Notification<serde_json::Value>),
-    Response(Response<serde_json::Value>),
-}
-
 /// A JSON-RPC Client over Websockets.
 ///
 /// ```no_run
@@ -112,7 +111,6 @@ impl Ws {
         S: Send + Sync + Stream<Item = WsStreamItem> + Sink<Message, Error = WsError> + Unpin,
     {
         let (sink, stream) = mpsc::unbounded();
-
         // Spawn the server
         WsServer::new(ws, stream).spawn();
 
@@ -134,13 +132,26 @@ impl Ws {
 
     /// Initializes a new WebSocket Client
     #[cfg(not(target_arch = "wasm32"))]
-    pub async fn connect(
-        url: impl tungstenite::client::IntoClientRequest + Unpin,
-    ) -> Result<Self, ClientError> {
+    pub async fn connect(url: impl IntoClientRequest + Unpin) -> Result<Self, ClientError> {
         let (ws, _) = connect_async(url).await?;
         Ok(Self::new(ws))
     }
 
+    /// Initializes a new WebSocket Client with authentication
+    #[cfg(not(target_arch = "wasm32"))]
+    pub async fn connect_with_auth(
+        uri: impl IntoClientRequest + Unpin,
+        auth: Authorization,
+    ) -> Result<Self, ClientError> {
+        let mut request: HttpRequest<()> = uri.into_client_request()?;
+
+        let mut auth_value = http::HeaderValue::from_str(&auth.to_string())?;
+        auth_value.set_sensitive(true);
+
+        request.headers_mut().insert(http::header::AUTHORIZATION, auth_value);
+        Self::connect(request).await
+    }
+
     fn send(&self, msg: Instruction) -> Result<(), ClientError> {
         self.instructions.unbounded_send(msg).map_err(to_client_error)
     }
@@ -169,19 +180,16 @@ impl JsonRpcClient for Ws {
         // send the data
         self.send(payload)?;
 
-        // wait for the response
-        let res = receiver.await?;
-
-        // in case the request itself has any errors
-        let res = res?;
+        // wait for the response (the request itself may have errors as well)
+        let res = receiver.await??;
 
         // parse it
-        Ok(serde_json::from_value(res)?)
+        Ok(serde_json::from_str(res.get())?)
     }
 }
 
 impl PubsubClient for Ws {
-    type NotificationStream = mpsc::UnboundedReceiver<serde_json::Value>;
+    type NotificationStream = mpsc::UnboundedReceiver<Box<RawValue>>;
 
     fn subscribe<T: Into<U256>>(&self, id: T) -> Result<Self::NotificationStream, ClientError> {
         let (sink, stream) = mpsc::unbounded();
@@ -309,28 +317,33 @@ where
     }
 
     async fn handle_text(&mut self, inner: String) -> Result<(), ClientError> {
-        match serde_json::from_str::<Incoming>(&inner) {
-            Err(_) => {}
-            Ok(Incoming::Response(resp)) => {
-                if let Some(request) = self.pending.remove(&resp.id) {
-                    if !request.is_canceled() {
-                        request.send(resp.data.into_result()).map_err(to_client_error)?;
-                    }
-                }
-            }
-            Ok(Incoming::Notification(notification)) => {
-                let id = notification.params.subscription;
-                if let Entry::Occupied(stream) = self.subscriptions.entry(id) {
-                    if let Err(err) = stream.get().unbounded_send(notification.params.result) {
-                        if err.is_disconnected() {
-                            // subscription channel was closed on the receiver end
-                            stream.remove();
-                        }
-                        return Err(to_client_error(err))
-                    }
-                }
+        let (id, result) = match serde_json::from_str(&inner)? {
+            Response::Success { id, result } => (id, Ok(result.to_owned())),
+            Response::Error { id, error } => (id, Err(error)),
+            Response::Notification { params, .. } => return self.handle_notification(params),
+        };
+
+        if let Some(request) = self.pending.remove(&id) {
+            if !request.is_canceled() {
+                request.send(result).map_err(to_client_error)?;
             }
         }
+
+        Ok(())
+    }
+
+    fn handle_notification(&mut self, params: Params<'_>) -> Result<(), ClientError> {
+        let id = params.subscription;
+        if let Entry::Occupied(stream) = self.subscriptions.entry(id) {
+            if let Err(err) = stream.get().unbounded_send(params.result.to_owned()) {
+                if err.is_disconnected() {
+                    // subscription channel was closed on the receiver end
+                    stream.remove();
+                }
+                return Err(to_client_error(err))
+            }
+        }
+
         Ok(())
     }
 
@@ -346,6 +359,7 @@ where
     async fn handle(&mut self, resp: Message) -> Result<(), ClientError> {
         match resp {
             Message::Text(inner) => self.handle_text(inner).await,
+            Message::Frame(_) => Ok(()), // Server is allowed to send Raw frames
             Message::Ping(inner) => self.handle_ping(inner).await,
             Message::Pong(_) => Ok(()), // Server is allowed to send unsolicited pongs.
             Message::Close(Some(frame)) => Err(ClientError::WsClosed(frame)),
@@ -442,6 +456,21 @@ pub enum ClientError {
     /// Something caused the websocket to close
     #[error("WebSocket connection closed unexpectedly")]
     UnexpectedClose,
+
+    /// Could not create an auth header for websocket handshake
+    #[error(transparent)]
+    #[cfg(not(target_arch = "wasm32"))]
+    WsAuth(#[from] http::header::InvalidHeaderValue),
+
+    /// Unable to create a valid Uri
+    #[error(transparent)]
+    #[cfg(not(target_arch = "wasm32"))]
+    UriError(#[from] http::uri::InvalidUri),
+
+    /// Unable to create a valid Request
+    #[error(transparent)]
+    #[cfg(not(target_arch = "wasm32"))]
+    RequestError(#[from] http::Error),
 }
 
 impl From<ClientError> for ProviderError {
@@ -484,11 +513,21 @@ mod tests {
         let mut blocks = Vec::new();
         for _ in 0..3 {
             let item = stream.next().await.unwrap();
-            let block = serde_json::from_value::<Block<TxHash>>(item).unwrap();
+            let block: Block<TxHash> = serde_json::from_str(item.get()).unwrap();
             blocks.push(block.number.unwrap_or_default().as_u64());
         }
 
         assert_eq!(sub_id, 1.into());
         assert_eq!(blocks, vec![1, 2, 3])
     }
+
+    #[tokio::test]
+    async fn deserialization_fails() {
+        let ganache = Ganache::new().block_time(1u64).spawn();
+        let (ws, _) = tokio_tungstenite::connect_async(ganache.ws_endpoint()).await.unwrap();
+        let malformed_data = String::from("not a valid message");
+        let (_, stream) = mpsc::unbounded();
+        let resp = WsServer::new(ws, stream).handle_text(malformed_data).await;
+        assert!(resp.is_err(), "Deserialization should not fail silently");
+    }
 }
